import contextlib
import os
import re
import shutil
import tarfile
import tempfile
from typing import Optional

from cargo2rpm.metadata import Metadata

from rust2rpm.cratesio import download_crate
from rust2rpm import log
from rust2rpm.patching import make_patches


LICENSE_FILE_PATTERN = re.compile(
    r"""
    COPYING(?:[.-].*)?|COPYRIGHT(?:[.-].*)?|
    EULA(?:[.-].*)?|[Ll]icen[cs]e|[Ll]icen[cs]e.*|
    (?:.*[.-])?(?:UN)?LICEN[CS]E(?:[.-].*)?|NOTICE(?:[.-].*)?|
    PATENTS(?:[.-].*)?|
    (?:agpl|l?gpl)[.-].*|CC-BY-.*|
    (?:AGPL|APACHE|BSD|GFDL|GNU|L?GPL|MIT|MPL|OFL)-.*[0-9].*
    """,
    re.VERBOSE,
)

LICENSE_EXCLUDE_DIRS = {
    "target",
    "vendor",
    "example",
    "examples",
    "_example",
    "_examples",
    "testdata",
    "_testdata",
    ".github",
    "tests",
    "test",
}

DOC_FILE_PATTERN = re.compile(
    r"""
    .*\.(?:md|markdown|mdown|mkdn|rst|txt)|AUTHORS|
    AUTHORS[.-].*|CONTRIBUTORS|CONTRIBUTORS[.-].*|README|
    README[.-].*|CHANGELOG|CHANGELOG[.-].*|TODO|TODO[.-].*
    """,
    re.IGNORECASE | re.VERBOSE,
)

DOC_FILE_EXCLUDES = re.compile(r"CMakeLists\.txt|versions\.txt|.*\.tpl|.*\.in")


def local_toml_file(toml_path: str) -> tuple[str, list[str], list[str]]:
    assert os.path.isfile(toml_path)
    assert os.path.basename(toml_path) == "Cargo.toml"

    parent = os.path.dirname(toml_path)
    doc_files = get_doc_files(parent)
    license_files = get_license_files(parent)

    return toml_path, doc_files, license_files


def local_cargo_dir(project_dir: str) -> tuple[str, list[str], list[str]]:
    assert os.path.isdir(project_dir)

    toml_path = os.path.join(project_dir, "Cargo.toml")
    doc_files = get_doc_files(project_dir)
    license_files = get_license_files(project_dir)

    return toml_path, doc_files, license_files


def parse_crate_file_name(path: str) -> tuple[str, str]:
    name, version = os.path.basename(path).removesuffix(".crate").rsplit("-", 1)
    return name, version


@contextlib.contextmanager
def files_from_crate(crate_path: str, crate_name: str, crate_version: str):
    """Unpacks crate_path and returns path to toml file, list of doc files, list of license files"""
    # -> tuple[str, list[str], list[str]]

    with tempfile.TemporaryDirectory() as tmpdir:
        target_dir = f"{tmpdir}/"

        with tarfile.open(crate_path, "r") as archive:
            for n in archive.getnames():
                if not os.path.abspath(os.path.join(target_dir, n)).startswith(target_dir):
                    raise Exception("Unsafe filenames!")
            archive.extractall(target_dir)

        toml_path = f"{tmpdir}/{crate_name}-{crate_version}/Cargo.toml"
        if not os.path.isfile(toml_path):
            raise IOError("Crate does not contain a Cargo.toml file.")

        root_path = f"{tmpdir}/{crate_name}-{crate_version}"
        doc_files = get_doc_files(root_path)
        license_files = get_license_files(root_path)

        yield toml_path, doc_files, license_files


def get_license_files(path: str) -> list[str]:
    """Heuristic match on file names to detect license files"""

    results: list[str] = []

    for root, dirs, files in os.walk(path, topdown=True):
        dirs[:] = [d for d in dirs if d not in LICENSE_EXCLUDE_DIRS]
        for f in files:
            if LICENSE_FILE_PATTERN.match(f):
                results.append(os.path.relpath(os.path.join(root, f), path))

    results.sort()
    return results


def get_doc_files(path: str) -> list[str]:
    """Heuristic match on file names to detect documentation files"""

    results: list[str] = []

    for root, dirs, files in os.walk(path, topdown=True):
        dirs[:] = []
        for f in files:
            if (
                DOC_FILE_PATTERN.fullmatch(f)
                and not LICENSE_FILE_PATTERN.fullmatch(f)
                and not DOC_FILE_EXCLUDES.fullmatch(f)
            ):
                results.append(os.path.relpath(os.path.join(root, f), path))

    results.sort()
    return results


def project_is_path(path: str) -> bool:
    return "/" in path or path in {".", ".."}


def guess_local_project_name_from_path(project: str) -> str:
    if os.path.isdir(project):
        dir_name = os.path.split(os.path.abspath(project))[1]
    else:
        dir_name = os.path.split(os.path.dirname(os.path.abspath(project)))[1]
    name = dir_name.rstrip("0123456789.").removesuffix("-")
    return name


def process_project_local(
    project: str, patch: bool, patch_foreign: bool
) -> tuple[str, tuple[Optional[str], Optional[str]], Metadata, list[str], list[str]]:
    if os.path.isdir(project):
        toml_path, doc_files, license_files = local_cargo_dir(project)
    else:
        toml_path, doc_files, license_files = local_toml_file(project)

    metadata = Metadata.from_cargo(toml_path)

    if len(metadata.packages) > 1:
        log.warn(f"More than one package in {toml_path!r} crate metadata.")
        log.warn("Skipping automatic creation of patches.")

        # fall back to the directory name for determining the name / version
        # of the project heuristically
        name = guess_local_project_name_from_path(project)

        log.warn(f"Falling back to {name!r} as the name of the project (based on the name of the containing folder).")
        diffs = (None, None)

    else:
        package = metadata.packages[0]
        name = package.name
        features = package.get_feature_names()
        diffs = make_patches(name, package.version, patch, patch_foreign, toml_path, features)

        # ensure metadata is up-to-date with changes from patches
        metadata = Metadata.from_cargo(toml_path)

    return name, diffs, metadata, doc_files, license_files


def process_project(
    project: str,
    version: Optional[str],
    patch: bool,
    patch_foreign: bool,
    store_crate: bool,
) -> tuple[str, tuple[Optional[str], Optional[str]], Metadata, list[str], list[str]]:
    if project_is_path(project) and not project.endswith(".crate"):
        # project points into unpacked sources:
        if store_crate:
            raise ValueError("The '--store-crate' / '-s' flag cannot be used for unpacked sources.")

        # process unpacked sources
        return process_project_local(project, patch, patch_foreign)

    if project_is_path(project) and project.endswith(".crate"):
        # project points at a local .crate file
        crate_file_path = project
        # determine name and version from the filename
        name, version = parse_crate_file_name(project)

    else:
        # project is just a crate name
        name = project
        # download .crate from crates.io;
        # set version to the latest stable version if it was not specified
        crate_file_path, version = download_crate(project, version)

    if store_crate:
        copy_target = os.path.join(os.getcwd(), os.path.basename(crate_file_path))

        if not (os.path.exists(copy_target) and os.path.samefile(crate_file_path, copy_target)):
            shutil.copy2(crate_file_path, copy_target)

    # process files from  a .crate archive
    with files_from_crate(crate_file_path, name, version) as (toml_path, doc_files, license_files):
        if not license_files:
            log.error(f"No license files detected in {name!r}.")

        metadata = Metadata.from_cargo(toml_path)

        if len(metadata.packages) > 1:
            log.error("Attempting to process a .crate file which contains a cargo workspace.")
            log.error("This mode of operation is unusual and not supported by rust2rpm.")
            raise ValueError("Failed to process invalid .crate file (cargo workspace)")

        package = metadata.packages[0]
        features = package.get_feature_names()
        diffs = make_patches(name, version, patch, patch_foreign, toml_path, features)

        # ensure metadata is up-to-date with changes from patches
        metadata = Metadata.from_cargo(toml_path)

    return name, diffs, metadata, doc_files, license_files
